# -*- coding: utf-8 -*-

from gui import GUI
import csv
import numpy as np
import matplotlib #This is only for getting the default colors
import gtk
import my #Some helper functions
#import Tkinter #This should not be required, but is needed for Py-Installer
#import FileDialog #This should not be required, but is needed for Py-Installer
#import scipy.linalg.cython_blas #This should not be required, but is needed for Py-Installer
#import scipy.linalg.cython_lapack #This should not be required, but is needed for Py-Installer
#import scipy.special._ufuncs_cxx #This should not be required, but is needed for Py-Installer
import statsmodels as sm
import statsmodels.api as sma
import statsmodels.formula.api as smf
import pandas as pd #For the DataFrame object
import warnings
#import cProfile

class GetGrowth(object):
    """ Main controller for the GetGrowth software.
        
        Performs calculations and serves as an interface to the GUI
    """
    
    #The GUI. This will also display the GUI
    _gui=None
    _times=None #Vector with the measured times
    _maxTime=1 #This can be used for setting an upper limit on the times (and corresponding values) which are returned. 0-1
    _values=None #Nx48 array with the measured values
    _toPlot=None #Boolean vector which determines if a well should be plotted and participate in calculations
    _wellColors=None #Nx4 array with the RGBA color to use for each well
    _plateDim=None #Tuplet with the dimensions of the plate
    _wellLabels=None #The labels for individual well (not groups)
    _inGroup=None #Boolean vector with True if a well is a member of a group
    _regressionLimits=None #48x2 array with the limits for the linear regression. NaN is used when the values haven't been set
    _regressionResults=None #list with the interpolation results for an individual well. NaN is used when the values haven't been set
    _delimiter=None #The delimiter used when loading the latest CSV file. This is also used when exporting, in order to maintain copmatibility
    
    _groupCounter=None #Counter for how many groups have been created. Used in the naming of groups
    _groupLabels=None #The labels for groups
    _toPlotGroups=None #Boolean vector which determines if a group should be plotted and participate in calculations
    _groupMembers=None #List of lists with the wells for each group
    _groupRegressionLimits=None #Nx2 array with the limits for the linear regression. NaN is used when the values haven't been set
    _groupRegressionResults=None #list with the interpolation results for a group. NaN is used when the values haven't been set
    
    #Properties       
    @property
    def timePoints(self):
        """ Array of timepoints for the measurements."""
        if self._maxTime==1:
            return self._times
        else:
            #Crop to maxTime
            return self._times[self._times<=(self._maxTime*self._times[-1])]
    @property
    def wellLabels(self):
        """ List of labels for the individual wells."""
        
        return self._wellLabels
     
    @property
    def values(self):
        """ Array with measurements for the wells."""
        
        if not self._values is None:
            if np.isinf(self._maxTime):
                return self._values
            else:
                #Crop to maxTime
                return self._values[self._times<=(self._maxTime*self._times[-1]),:]
        else:
            return None

    @property
    def toPlot(self):
        """ Boolean array with the wells to plot."""
        
        return self._toPlot
    
    @property
    def colors(self):
        """ Array with colors for the individual wells."""
        
        return self._wellColors
        
    @property
    def regressionLimits(self):
        """ Nx2 array with regression limits for the individual wells."""
        
        return self._regressionLimits
         
    @property
    def groupLabels(self):
        """ Array with labels for the groups."""
        
        return self._groupLabels
        
    @property
    def toPlotGroups(self):
        """ Boolean array with the groups to plot."""
        
        return self._toPlotGroups
        
    @property
    def inGroup(self):
        """ Boolean array which determines if a well participates in a group."""
        
        return self._inGroup
        
    def __init__(self):
        self.initVariables()
        
    def showGUI(self):
        """ Load and show the GUI."""
        
        self._gui=GUI(self)
        
        #Show GUI and give control over to the main GUI loop
        self._gui.show()
        
    def setMaxTime(self,value):
        """ Set the maximal time factor."""
        self._maxTime=value
    
    def initVariables(self):
        """ (Re)set all variables to their default values """
        
        self._toPlot=np.ones(48,dtype=bool)
        self._inGroup=np.zeros(48,dtype=bool)
        self._regressionLimits=np.empty((48,2))
        self._regressionLimits[:]=np.nan
        self._regressionResults=[None]*48
        self._wellLabels=["A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8",
                        "B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8",
                        "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8",
                        "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8",
                        "E1", "E2", "E3", "E4", "E5", "E6", "E7", "E8",
                        "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8"]
        self._times=None
        self._values=None
        self._wellColors=matplotlib.cm.gist_rainbow(np.linspace(0,1,48))
        self._plateDim=(6,8)
        self._groupCounter=1
        self._groupLabels=None
        self._toPlotGroups=None
        self._groupMembers=None
        self._groupRegressionLimits=None
        self._groupRegressionResults=None
        
    def fitGroup(self,index):
        """ Fits a Random intercept, random slope-model to the curves in a group. Note
            that the log of the values will be used and that the fitting will be done in
            the interval specified by regressionLimits.
        """
        
        startT=self._groupRegressionLimits[index,0]
        endT=self._groupRegressionLimits[index,1]
        
        if not np.isnan(startT) and not np.isnan(endT):
            #Get the data. 
            I=np.where(np.bitwise_and(self._times>=startT, self._times<=endT))[0]
            
            p=np.log(self._values[I,:])
            
            members=self._groupMembers[index]
            
            data=np.zeros((0,3))
            
            
            for i in range(0,len(members)):
                toAdd=np.column_stack((self._times[I],p[:,members[i]],i*np.ones((len(I),1))))
                data=np.row_stack((data,toAdd))
                
                #Also fit a line to each member. This is not used here, but is printed when saving
                self._regressionLimits[members[i],0]=startT
                self._regressionLimits[members[i],1]=endT
                self.fitLine(members[i])
            
            result=self._MELR(data)
            ransacResult=self._fitRANSAC(data[:,0],data[:,1],data[:,2],[self._regressionResults[i] for i in members])
            
            if not ransacResult is None:
                result['ransac_intercept']=ransacResult['intercept']
                result['ransac_slope']=ransacResult['slope']
                result['ransac_confInterval']=ransacResult['confInterval']
                result['ransac_upperLegend']=ransacResult['upperLegend']
                result['ransac_lowerLegend']=ransacResult['lowerLegend']
            else:
                result['ransac_intercept']=np.nan
                result['ransac_slope']=np.nan
                result['ransac_rsquared']=np.nan
                result['ransac_confInterval']=[np.nan,np.nan]
                result['ransac_upperLegend']='No model found'
                result['ransac_lowerLegend']='No model found'
            self._groupRegressionResults[index]=result
        
    def _fitRANSAC(self,t,y,types=None,prevFit=None):
        """ Performs a robust linear regression using the RANSAC algorithm.
        
            types:      the line index for each value in y (if for a group)
            prevFit:    a result structure or a list of result structures (if for a group) 
        
            Returns a results dictionary
        """
        
        #RANSAC parameters
        proportionToInclude=0.7
        minProportionForModel=0.7
        nIterations=500
        
        #To be able to run same code for groups
        if types is None:
            singleFit=True
            types=np.zeros(len(t))
        else:
            singleFit=False
            
        if isinstance(prevFit,dict):
            prevFit=[prevFit]
        
        #True if a point is in the final fitting
        inPoints=np.zeros(len(t),dtype=bool)
        
        #Do this for each line
        for i in range(0,len(prevFit)):
            I=np.where(types==i)[0]
            
            #First get the points from the previous fit
            oldY=prevFit[i]['intercept']+t[I]*prevFit[i]['slope']
            
            #Ignore points which have larger residuals than this
            maxRes=np.percentile(np.sort(np.abs(oldY-y[I])),proportionToInclude*100)
            
            for j in range(0,nIterations):
                #Select two random points in the range
                limits=np.random.choice(len(I),2,replace=False)
                x0=min(limits)
                x1=max(limits)
                
                #Calculate the line between the two points
                slope=(y[I[x1]]-y[I[x0]])/(t[I[x1]]-t[I[x0]])
                intercept=y[I[x0]]-t[I[x0]]*slope
                
                #Calculate the predicted points for the line
                p=intercept+slope*t[I]
                
                goodFits=np.abs(p-y[I])<=maxRes
                if sum(goodFits)>=minProportionForModel*len(I):
                    inPoints[I]=np.bitwise_or(inPoints[I],goodFits)
        
        #Perform linear regression on all inPoints
        if any(inPoints):
            if singleFit is False:
                return self._MELR(np.column_stack((t[inPoints],y[inPoints],types[inPoints])))
            else:
                return self._singleLR(t[inPoints],y[inPoints])
        else:
            return None
            
    def _MELR(self,data):
        """ Perform mixed-effects regression for a group.
        
            This is split from fitGroup to be able to use it from fitRANSAC
            
            Returns a dictionary with the results
        """
        
        df=pd.DataFrame(data,columns=['TIME', 'OD', 'TYPE'])
        
        #To set that there should be no interaction term
        free=sm.regression.mixed_linear_model.MixedLMParams.from_components(np.ones(2),np.eye(2))
        md = smf.mixedlm('OD ~ TIME', df, groups=df['TYPE'],re_formula='~TIME')
        
        #This warning is shown if the variance in the random slope is too small. It doesn't matter so hide it
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            results = md.fit(free=free)
        
        #Dictionary with the results
        result={}
        result['intercept']=results.params['Intercept']
        result['slope']=results.params['TIME']
        confInt=results.conf_int(0.05)
        result['confInterval']=[confInt[0]['TIME'],confInt[1]['TIME']]
        result['upperLegend']='µ: ' + str(round(result['slope'],3)) + ' [' + str(round(result['confInterval'][0],3)) + ', ' + str(round(result['confInterval'][1],3)) + ']'
        result['lowerLegend']=None
        
        return result
            
    def _singleLR(self,x,y):
        """ Perform linear regression on a single line.
        
            This is split from fitLine to be able to use it from fitRANSAC
            
            Returns a dictionary with the results
        """
        
        model=sm.regression.linear_model.OLS(y,sm.regression.linear_model.add_constant(x))
        results = model.fit()
        
        #Dictionary with the results
        result={}
        result['intercept']=results.params[0]
        result['slope']=results.params[1]
        result['rsquared']=results.rsquared
        result['confInterval']=results.conf_int(0.05)[1,:]
        result['upperLegend']='µ: ' + str(round(result['slope'],3)) + ' [' + str(round(result['confInterval'][0],3)) + ', ' + str(round(result['confInterval'][1],3)) + ']'
        result['lowerLegend']='R-sqr: ' + str(round(result['rsquared'],3))
        
        return result
        
     
    def fitLine(self,index,noRANSAC=False):
        """ Fits a line to a curve and returns the fitting.
        
            Note that the log of the values will be used and that the fitting
            will be done in the interval specified by regressionLimits. This also
            calls on fitRANSAC for a robust fitting.
        """
        
        startT=self._regressionLimits[index,0]
        endT=self._regressionLimits[index,1]
        
        if not np.isnan(startT) and not np.isnan(endT):
            #Get the data
            I=np.where(np.bitwise_and(self._times>=startT, self._times<=endT))
            
            p=np.log(self._values[I,index].flatten(1))
            t=self._times[I]
            
            results=self._singleLR(t,p)
            if noRANSAC is False:
                ransacResults=self._fitRANSAC(t,p,None,results)
                if not ransacResults is None:
                    results['ransac_intercept']=ransacResults['intercept']
                    results['ransac_slope']=ransacResults['slope']
                    results['ransac_rsquared']=ransacResults['rsquared']
                    results['ransac_confInterval']=ransacResults['confInterval']
                    results['ransac_upperLegend']=ransacResults['upperLegend']
                    results['ransac_lowerLegend']=ransacResults['lowerLegend']
                else:
                    results['ransac_intercept']=np.nan
                    results['ransac_slope']=np.nan
                    results['ransac_rsquared']=np.nan
                    results['ransac_confInterval']=[np.nan,np.nan]
                    results['ransac_upperLegend']='No model found'
                    results['ransac_lowerLegend']='No model found'

            self._regressionResults[index]=results
            
    def getRegressionResults(self,indexes):
        """ Create a list of regression results from the indexes."""
        
        tempList=[]
        
        for index in indexes:
            tempList.append(self._regressionResults[index])
       
        return tempList
        
    def setWellLabel(self,index,value):
        """ Set the label of a well."""
        
        if value.count(',')==0 and value.count(';')==0:
            self._wellLabels[index]=value
    
    def setGroupLabel(self,index,value):
        """ Set the label of a group."""
        
        if value.count(',')==0 and value.count(';')==0:
            self._groupLabels[index]=value
    
    def getGroupRegressionResults(self,indexes):
        """ Create a list of regression results from the group indexes."""
        
        tempList=[]
        
        for index in indexes:
            tempList.append(self._groupRegressionResults[index])
       
        return tempList
        
    def saveFile(self,fileName):
        """ Save results to a CSV file."""
        d=self._delimiter
        
        with open(fileName, "w") as csv:
            #Get groups to print
            I=np.where(self.toPlotGroups)[0]
            
            if len(I)>0:
                csv.write('***Groups' + d*8 + '\n')
                csv.write(d*2 + 'mu (1/h)' + d + '95% CI' + d + 'R-sqr' + d + 'mu (1/h) [RANSAC]' + d + '95% CI [RANSAC]' + d + 'R-sqr [RANSAC]' + d + 'Time interval (h)\n')
                
                for index in I:
                    result=self.getGroupRegressionResults([index])[0]
                    limits=self.getGroupRegressionLimits(index)[0]
                    
                    if not result is None:   
                        u='%.3f' % round(result['slope'],3)
                        lCI='%.3f' % round(result['confInterval'][0],3)
                        uCI='%.3f' % round(result['confInterval'][1],3)
                        CI='[' + lCI + ' - ' + uCI + ']'
                        u_r='%.3f' % round(result['ransac_slope'],3)
                        lCI_r='%.3f' % round(result['ransac_confInterval'][0],3)
                        uCI_r='%.3f' % round(result['ransac_confInterval'][1],3)
                        CI_r='[' + lCI_r + ' - ' + uCI_r + ']'
                        lTI='%.3f' % round(limits[0],3)
                        uTI='%.3f' % round(limits[1],3)
                        TI='[' + lTI + ' - ' + uTI + ']'
                        
                        csv.write(self.groupLabels[index] + d*2 + u + d + CI + d*2 + u_r + d + CI_r + d*2 + TI + '\n') 
                    else:
                        csv.write(self.groupLabels[index] + d*8 + '\n')
                    
                    #Then print each well in the group
                    J=np.intersect1d(np.where(self.toPlot)[0],self.getGroupMembers(index))
                    for index in J:
                        result=self.getRegressionResults([index])[0]
                        limits=self.regressionLimits
                        if not result is None:
                            u='%.3f' % round(result['slope'],3)
                            lCI='%.3f' % round(result['confInterval'][0],3)
                            uCI='%.3f' % round(result['confInterval'][1],3)
                            CI='[' + lCI + ' - ' + uCI + ']'
                            R2='%.3f' % round(result['rsquared'],3)
                            u_r='%.3f' % round(result['ransac_slope'],3)
                            lCI_r='%.3f' % round(result['ransac_confInterval'][0],3)
                            uCI_r='%.3f' % round(result['ransac_confInterval'][1],3)
                            CI_r='[' + lCI_r + ' - ' + uCI_r + ']'
                            R2_r='%.3f' % round(result['ransac_rsquared'],3)
    
                            csv.write(d + self.wellLabels[index] + d + u + d + CI + d + R2 + d + u_r + d + CI_r + d + R2_r + d +'\n') 
                        else:
                            csv.write(d + self.wellLabels[index] + d*7 + '\n') 
                            
                    csv.write(d*8 + '\n')
                
            #Get wells to print
            I=np.where(np.bitwise_and(self.toPlot,~self.inGroup))[0]
            if len(I)>0:
                csv.write('***Wells' + d*5 + '\n')
                csv.write(d + 'mu (1/h)' + d + '95% CI' + d + 'R-sqr' + d + 'mu (1/h) [RANSAC]' + d + '95% CI [RANSAC]' + d + 'R-sqr [RANSAC]' + d + 'Time interval (h)' + d + '\n') 
                for index in I:
                    result=self.getRegressionResults([index])[0]
                    limits=self.regressionLimits
                    if not result is None:
                        u='%.3f' % round(result['slope'],3)
                        lCI='%.3f' % round(result['confInterval'][0],3)
                        uCI='%.3f' % round(result['confInterval'][1],3)
                        CI='[' + lCI + ' - ' + uCI + ']'
                        R2='%.3f' % round(result['rsquared'],3)
                        u_r='%.3f' % round(result['ransac_slope'],3)
                        lCI_r='%.3f' % round(result['ransac_confInterval'][0],3)
                        uCI_r='%.3f' % round(result['ransac_confInterval'][1],3)
                        CI_r='[' + lCI_r + ' - ' + uCI_r + ']'
                        R2_r='%.3f' % round(result['ransac_rsquared'],3)
                        lTI='%.3f' % round(limits[index,0],3)
                        uTI='%.3f' % round(limits[index,1],3)
                        TI='[' + lTI + ' - ' + uTI + ']'

                        csv.write(self.wellLabels[index] + d + u + d + CI + d + R2 + d + u_r + d + CI_r + d + R2_r + d + TI + d +'\n') 
                    else:
                        csv.write(self.wellLabels[index] + d*8 + '\n')
     
    def loadFile(self,fileName):
        """ Load a CSV file."""
        
        #Clear any current variables and set them to their default values
        self.initVariables()
        
        #The issue here is that some Biolector files are semicolon separated with
        #decimal comma and some are comma separated with decimal dot. First try to figure
        #out the delimiter by reading the first line and checking for existance of comma or semicolon
        self._delimiter=None
        with open(fileName, 'rb') as csvfile:
            row=csvfile.next()
            commas=row.count(',')
            semicolons=row.count(';')
            
            if commas==1 and semicolons==0:
                self._delimiter=','
            if semicolons==1 and commas==0:
                self._delimiter=';'
            
        #Could not get delimiter, return without error message for now
        if self._delimiter is None:
            return
        
        with open(fileName, 'rb') as csvfile:
            csvreader = csv.reader(csvfile,delimiter=self._delimiter)
            
            #First search to the row where the third row is "TIME [h] ->"
            breakEarly=False
            for row in csvreader:
                if breakEarly is False:
                    if len(row)>=4:
                        if row[3]=='TIME [h] ->':
                            #The current format uses "," as decimal point. This should be checked more carefully
                            self._times=np.zeros(len(row)-4)
                            for i in range(4,len(row)):
                                self._times[i-4]=float(row[i].replace(',','.'))
                                
                            #The Biolector file format may sometimes miss the last measured value for
                            #some wells. This is probably when the program has been aborted prematurely.
                            #Trim all measurements to the shortest sample
                            
                            #Then read the remainder of the values into the value matrix. Only read 48 rows for now (first channel)
                            self._values=np.empty([len(self._times),0])
                            counter=0
                            for row in csvreader:
                                counter=counter+1
                                if counter<=48:
                                    #First convert the values in the row
                                    newRow=np.zeros(len(row)-4)
                                    for i in range(4,len(row)):
                                            newRow[i-4]=float(row[i].replace(',','.'))
                                    
                                    #This means that this row is shorter then the others. Trim old values        
                                    if len(newRow)<len(self._times):
                                        self._times=self._times[0:len(newRow)]
                                        self._values=self._values[0:len(newRow),:]
                                    self._values=my.cat(self._values,newRow[0:len(self._times)],1)
                                else:
                                    breakEarly=True
                                    break
        
        #Normalize values to start at 0. This is problematic if the real OD at t0 is a significant fraction of the
        #OD used for fitting. Add 0.5% of max value for less noise at low levels
        for i in range(0,self._values.shape[1]):
            self._values[:,i]=self._values[:,i]-min(self._values[:,i])
            self._values[:,i]=self._values[:,i]+0.005*max(self._values[:,i])

    def toggleToPlot(self,index):
        """ Toggle whether to plot a well or not."""
        
        self._toPlot[index]=np.logical_not(self._toPlot[index])

    def toggleToPlotGroup(self,index):
        """ Toggle whether to plot a group or not."""
        
        self._toPlotGroups[index]=np.logical_not(self._toPlotGroups[index])
        
    def addToGroup(self,groupIndex,wellIndex):
        """ Add a single well to a group."""
        
        if self._inGroup[wellIndex]==False:
            self._inGroup[wellIndex]=True
            self._groupMembers[groupIndex]=my.cat(self._groupMembers[groupIndex],np.array([wellIndex]))
        
    def getGroupMembers(self,groupIndex):
        """ Get the wells in a given group."""
        return self._groupMembers[groupIndex]
        
    def getGroupRegressionLimits(self,groupIndex):
        """ Get the regression limit for a group. np.nan is used if not set."""
        return self._groupRegressionLimits[groupIndex,:][None,:]
        
    def getGroupColor(self,groupIndex):
        """ Get the color for a group. This is currently the color for the first
            member of the group, but that may change.
            
            Returns both an RGBA array and an gdk.Color object
        """

        members=self._groupMembers[groupIndex]
        if len(members)>0:
            color=self._wellColors[members[0],:][None,:]
            return (color,gtk.gdk.Color(color[0,0],color[0,1],color[0,2]))
        else:
            return (None, None)
                
    def newGroup(self):
        """ Create a new group."""
        if not self._groupLabels is None:
            self._groupLabels.append('Group '+ str(self._groupCounter))
            self._toPlotGroups.append(True)
            self._groupMembers.append(np.zeros(0))
            nan=np.empty((1,2))
            nan[:]=np.nan
            self._groupRegressionLimits=my.cat(self._groupRegressionLimits,nan,0)
            self._groupRegressionResults.append(None)
            
        else:
            self._groupLabels=['Group '+ str(self._groupCounter)]
            self._toPlotGroups=[True]
            self._groupMembers=[np.zeros(0)]
            self._groupRegressionLimits=np.empty((1,2))
            self._groupRegressionLimits[:]=np.nan
            self._groupRegressionResults=[None]
            
        self._groupCounter=self._groupCounter+1
        
    def toggleMultipleToPlot(self,index=None,type=None):
        """ Toggle whether to plot a whole row/column or not.
        
            type=0: index is a row
            type=1: index is a column
            
            index is None: toggle all wells
        """
        
        #To convert between indexes first make a boolean matrix
        mat=np.zeros(self._plateDim)
        
        if not index is None:
            #Toggle row
            if type==0:
                mat[index,:]=True
            else:
                mat[:,index]=True
        else:
            mat[:,:]=True #Toggle all
            
        #Get indexes to toggle
        I=np.where(mat)
        J=np.ravel_multi_index(I, self._plateDim)
        
        if np.any(self._toPlot[J]):
            self._toPlot[J]=False
        else:
            self._toPlot[J]=True
        
if __name__ == "__main__":
    """ Run the program."""
    GetGrowth().showGUI()



import numpy as np
import os
import sys
import time

global timeDict
timeDict={}

global ct

def start_clock():
    global ct
    ct=time.time()
    
def stop_clock(tag):
    global timeDict
    global ct
    if tag in timeDict:
        timeDict[tag]=timeDict[tag]+time.time()-ct
    else:
        timeDict[tag]=time.time()-ct
        
    
def print_clock():
    global timeDict
    for tag in timeDict.keys():
        print(tag + ' ' + str(timeDict[tag]) +'\n')
    

def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)

def cat(A,B,dim=0):
    """ The purpose of this function is to expand on the functionality of the numpy
        concatenate function to allow for concatenation of one dimensional arrays and
        None values.
    """
    
    if A is None:
        return B
    
    if B is None:
        return A
        
    if len(A)==0:
        return B
    
    if len(B)==0:
        return A
        
    sA=A.shape
    sB=B.shape
    
    if len(sA)>1 and len(sB)>1:
        return np.concatenate((A,B),dim)
        
    #If both of them are one dimensional
    if len(sA)==1 and len(sB)==1:
        #Dim doesn't matter in this case
        return np.concatenate((A,B),dim)
        
    #If one of them is one dimensional
    if len(sA)==1:
        if dim==0:
            A=A[None,:]
        else:
            A=A[:,None]
    
    if len(sB)==1:
        if dim==0:
            B=B[None,:]
        else:
            B=B[:,None]
            
    return np.concatenate((A,B),dim)


from setuptools import setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(name='getgrowth',
      version='0.2.5',
      description='Determining growth rates from Biolector experiments',
      url='http://github.com/rasmusagren/getgrowth',
      author='Rasmus Agren',
      author_email='rasmus@alembic.se',
      license='GNU General Public License v2.0',
      packages=['getgrowth'],
      include_package_data=True,
      install_requires=[
          'pandas>=0.17.0',
          'numpy>=1.8',
          'matplotlib>=1.5',
          'pygtk>=2.24.0',
          'statsmodels>=0.6.1'],
      zip_safe=False)


# -*- coding: utf-8 -*-
import gtk
import numpy as np
import os
import subprocess
import matplotlib
matplotlib.use('GTKAgg')

from matplotlib.figure import Figure
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas

import copy
import time
import my #My helper functions for stuff

class GUI(object):

    """ The GUI."""
    
    #The controlling object
    _controller=None
    
    #Widgets which will be referred to in the callbacks
    _w={}
    
    #Figures to draw curves on
    _uFig=None
    _LFig=None
    _uCanvas=None
    _lCanvas=None
    
    #Default background colors
    _defaultGray=None
    _defaultWhite=None
    
    _addingToGroup=None #Used when adding wells to a group
    _fittingRegression=None #Used when fitting regression lines
    _currentRegressionIsGroup=None
    _currentRegressionIndex=None
    
    #This is the number of the item currently being used for fitting regression lines
    #If there are N groups and M individual wells 0 to N-1 refer to groups and N to N+(M-1)
    #refer to wells
    _currentRegressionNumber=None
    
    def __init__(self,controller):
                    
        """Construct the GUI from the Glade file."""
        
        self._controller=controller #This contains the controlling object
        builder=gtk.Builder()
        builder.add_from_file(os.path.join(os.path.dirname(__file__),'gui.glade'))
        
        #Load all widgets which will be referred to in the callbacks       
        self._w['window']=builder.get_object('window')
        self._w['uFig']=builder.get_object('uFig')
        self._w['lFig']=builder.get_object('lFig')
        self._w['wellListStore']=builder.get_object('wellListStore')
        self._w['treeStore']=builder.get_object('treeStore')
        self._w['treeView']=builder.get_object('treeView')
        self._w['wellTreeView']=builder.get_object('wellTreeView')
        self._w['newGroup']=builder.get_object('newGroup')
        self._w['addToGroup']=builder.get_object('addToGroup')
        self._w['fitRegression']=builder.get_object('fitRegression')
        self._w['prevRegression']=builder.get_object('prevRegression')
        self._w['nextRegression']=builder.get_object('nextRegression')
        self._w['xScale']=builder.get_object('xScale')
        self._w['regressionLabel']=builder.get_object('regressionLabel')
        
        builder.connect_signals(self)
        
        #For setting the range of the slider
        self._w['xScale'].set_range(0.1,1)
        self._w['xScale'].set_value(1)
        
        #Figures for plotting
        self._uFig=Figure(figsize=(5,4),dpi=100)
        self._lFig=Figure(figsize=(5,4),dpi=100)
        
        self._uCanvas=FigureCanvas(self._uFig)
        self._uCanvas.set_size_request(800,300)
        self._lCanvas=FigureCanvas(self._lFig)
        self._lCanvas.set_size_request(800,300)
        self._lCanvas.mpl_connect('button_press_event',self.on_lCanvas_clicked)
        
        self._w['uFig'].add_with_viewport(self._uCanvas)
        self._w['lFig'].add_with_viewport(self._lCanvas)
        
        #Populate the wellListStore
        self._defaultGray=gtk.gdk.color_parse('#C9C9C9')
        self._defaultWhite=gtk.gdk.color_parse('#FFFFFF')
        
        gray=self._defaultGray
        white=self._defaultWhite
        self._w['wellListStore'].append(["↓", "↓", "↓", "↓", "↓", "↓", "↓", "↓",white,white,white,white,white,white,white,white,"○"])
        self._w['wellListStore'].append(["A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        self._w['wellListStore'].append(["B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        self._w['wellListStore'].append(["C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        self._w['wellListStore'].append(["D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        self._w['wellListStore'].append(["E1", "E2", "E3", "E4", "E5", "E6", "E7", "E8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        self._w['wellListStore'].append(["F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8",gray,gray,gray,gray,gray,gray,gray,gray,"→"])
        
        self._fittingRegression=False
        self._currentRegressionIsGroup=False
        self._currentRegressionIndex=None
        
        #Disable interface until some file has been loaded
        self.changeEnabled(False)

    def show(self):
        """ Show the GUI and enter main loop."""

        self._w['window'].show_all()
        gtk.timeout_add(100,gtk.main_iteration_do,True)
        gtk.idle_add(gtk.main_iteration_do,True)
        gtk.main()
        
    def plotData(self,x=0,y=0,colors=None,dashed=None,intervals=None,regressionResults=None,target=0):
        """ Plot data on the upper or lower figures.
        
            x:          N array with the time points
            y:          NxM array where N is the sample points and M is the samples
            colors:     Mx4 array in RGBA format
            dashed:     boolean vector with True if a sample should be plotted with
                        dashed line
            intervals:  Mx2 array with the limits for the linear regression. np.nan
                        is used if no interval is set
            """
            
        #Note that this is ignored if no color is supplied. Should be changed at some point
        if dashed is None:
            dashed=np.zeros(y.shape[1],dtype=bool)
        if intervals is None:
            intervals=np.zeros((y.shape[1],2))
            intervals[:]=np.nan
            
        if regressionResults is None:
            regressionResults=[None]*y.shape[1]
        
        #If target is upper
        if target is 0:
            ax=self._uFig.add_subplot(111)
        else:
            ax=self._lFig.add_subplot(111)
        
        ax.set_xlim(x.min(), x.max())
        ax.set_ylim(y.min(), y.max())
        
        #This is to be able to use a LineCollection. This was a later addition and the
        #arguments for the function could be changed to get around this
            
        #First plot all solid lines
        if np.any(~dashed):
            segs=np.zeros((sum(~dashed), len(x), 2), float)
            segs[:,:,0]=x
            segs[:,:,1]=y.T[~dashed,:]

            lineCollection=matplotlib.collections.LineCollection(segs,colors=colors[~dashed,:])
        
            ax.add_collection(lineCollection)
        
        #Then dashed lines
        if np.any(dashed):
            segs=np.zeros((sum(dashed), len(x), 2), float)
            segs[:,:,0]=x
            segs[:,:,1]=y.T[dashed,:]

            lineCollection=matplotlib.collections.LineCollection(segs,colors=colors[dashed,:],linestyles='dotted')
        
            ax.add_collection(lineCollection)
            
        #Then add interval lines and regression results. This could also have been done
        #by a LineCollection, but since this is rather fast I keep it in a loop for now
        for interval,regressionResult in zip(intervals,regressionResults):
            
            #Draw vertical lines to indicate the regression limits
            if np.isnan(interval[0])==False:
                #Get the value for the non-dashed line at this point. Note that this
                #assumes that there will always be at least one non-dashed line when drawing
                #these marks
                refY=np.interp(interval[0],x,y[:,np.where(~dashed)[0][0]])
                ax.plot([interval[0],interval[0]],[refY-0.05*y.max(),refY+0.05*y.max()],linewidth=1,color=[0,0,0])
            if np.isnan(interval[1])==False:
                refY=np.interp(interval[1],x,y[:,np.where(~dashed)[0][0]])
                ax.plot([interval[1],interval[1]],[refY-0.05*y.max(),refY+0.05*y.max()],linewidth=1,color=[0,0,0])
            
            #Draw the regression information
            if not regressionResult is None:
                if target==0:
                    #Left part
                    rx=np.linspace(0,interval[0])
                    
                    if not np.isnan(regressionResult['ransac_intercept']):
                        ry=np.exp(regressionResult['ransac_intercept'])*np.power(np.exp(regressionResult['ransac_slope']),rx)
                        ax.plot(rx,ry,linewidth=1,color=[0.5,0.5,0.5],linestyle='-.')
                    ry=np.exp(regressionResult['intercept'])*np.power(np.exp(regressionResult['slope']),rx)
                    ax.plot(rx,ry,linewidth=1,color=[0,0,0],linestyle='-.')
                    
                    #Middle part
                    rx=np.linspace(interval[0],interval[1])
                    
                    if not regressionResult['ransac_intercept'] is None:
                        ry=np.exp(regressionResult['ransac_intercept'])*np.power(np.exp(regressionResult['ransac_slope']),rx)
                        ax.plot(rx,ry,linewidth=1,color=[0.5,0.5,0.5])
                    ry=np.exp(regressionResult['intercept'])*np.power(np.exp(regressionResult['slope']),rx)
                    ax.plot(rx,ry,linewidth=1,color=[0,0,0])
                    
                    #Right part
                    rx=np.linspace(interval[1],x.max())
                    
                    if not np.isnan(regressionResult['ransac_intercept']):
                        ry=np.exp(regressionResult['ransac_intercept'])*np.power(np.exp(regressionResult['ransac_slope']),rx)
                        ax.plot(rx,ry,linewidth=1,color=[0.5,0.5,0.5],linestyle='-.')
                    ry=np.exp(regressionResult['intercept'])*np.power(np.exp(regressionResult['slope']),rx)
                    ax.plot(rx,ry,linewidth=1,color=[0,0,0],linestyle='-.')
                    
                    if not regressionResult['upperLegend'] is None:
                        ax.text(0.02,0.9,regressionResult['upperLegend'] + ' (Normal)',fontsize=10,color=[0,0,0],transform = ax.transAxes)
                    if not regressionResult['ransac_upperLegend'] is None:
                        ax.text(0.02,0.82,regressionResult['ransac_upperLegend'] + ' (RANSAC)',fontsize=10,color=[0.5,0.5,0.5],transform = ax.transAxes)
                else:
                    #Left part
                    if not np.isnan(regressionResult['ransac_intercept']):
                        ry=[regressionResult['ransac_intercept'], regressionResult['ransac_intercept']+interval[0]*regressionResult['ransac_slope']]
                        ax.plot([0,interval[0]],ry,linewidth=1,color=[0.5,0.5,0.5],linestyle='-.')
                    ry=[regressionResult['intercept'], regressionResult['intercept']+interval[0]*regressionResult['slope']]
                    ax.plot([0,interval[0]],ry,linewidth=1,color=[0,0,0],linestyle='-.')
                        
                    #Middle part
                    if not np.isnan(regressionResult['ransac_intercept']):
                        ry=[regressionResult['ransac_intercept']+interval[0]*regressionResult['ransac_slope'], regressionResult['ransac_intercept']+interval[1]*regressionResult['ransac_slope']]
                        ax.plot(interval,ry,linewidth=1,color=[0.5,0.5,0.5])
                    ry=[regressionResult['intercept']+interval[0]*regressionResult['slope'], regressionResult['intercept']+interval[1]*regressionResult['slope']]
                    ax.plot(interval,ry,linewidth=1,color=[0,0,0])
                    
                    #Right part
                    if not np.isnan(regressionResult['ransac_intercept']):
                        ry=[regressionResult['ransac_intercept']+interval[1]*regressionResult['ransac_slope'], regressionResult['ransac_intercept']+x.max()*regressionResult['ransac_slope']]
                        ax.plot([interval[1],x.max()],ry,linewidth=1,color=[0.5,0.5,0.5],linestyle='-.')
                    ry=[regressionResult['intercept']+interval[1]*regressionResult['slope'], regressionResult['intercept']+x.max()*regressionResult['slope']]
                    ax.plot([interval[1],x.max()],ry,linewidth=1,color=[0,0,0],linestyle='-.')
                    
                    if not regressionResult['lowerLegend'] is None:
                        ax.text(0.02,0.9,regressionResult['lowerLegend'] + ' (Normal)',fontsize=10,transform = ax.transAxes)
                    if not regressionResult['ransac_lowerLegend'] is None:
                        ax.text(0.02,0.82,regressionResult['ransac_lowerLegend'] + ' (RANSAC)',fontsize=10,color=[0.5,0.5,0.5],transform = ax.transAxes)
        
        if target==0:
            ax.set_ylabel('original data')
        else:
            ax.set_ylabel('log-transformed data')
                                   
    def on_xScale_value_changed(self,widget,data=None):
        """ When the x-axis limit has been changed by the slider."""
        
        self._controller.setMaxTime(widget.get_value())
            
        #Re-scale x-axis. This check is because this is called once when the GUI is loaded, which is before the canvas exists
        if not self._uCanvas is None:
            if self._fittingRegression is False:
                #Clear figures
                self.clearFigures()
        
                #First plot all groups
                self.plotGroups(np.where(self._controller.toPlotGroups)[0])
                
                #Then add the individual wells
                self.plotWells(np.where(np.logical_and(self._controller.inGroup==False, self._controller.toPlot==True))[0])
                
                #Redraw figures
                self.redrawFigures()
            else:
                #If it's currently drawing a regression
                self.drawCurrentRegression()
            
    def on_menuSave_activate(self,widget,data=None):
        """ Export results to CSV file."""
        
        #Open save as dialog
        chooser=gtk.FileChooserDialog("Save File...", self._w['window'],
                                    gtk.FILE_CHOOSER_ACTION_SAVE,
                                    (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
                                    gtk.STOCK_SAVE, gtk.RESPONSE_OK))
                                    
        filter=gtk.FileFilter()
        filter.set_name("CSV files")
        filter.add_pattern("*.csv")
        chooser.add_filter(filter)
        
        chooser.set_current_name('GetGrowth results.csv')
    
        chooser.run()
        fileName=chooser.get_filename()
        chooser.destroy()
        
        if fileName is None:
            return
            
        #Append .csv if needed
        if not fileName[-4:].upper()=='.CSV':
            fileName=fileName+'.csv'
            
        self._controller.saveFile(fileName)
    
    def on_menuSample_activate(self,widget,data=None):
        """ Load a sample file."""
        
        self._controller.loadFile(os.path.join(os.path.dirname(__file__),'biolector.csv'))
            
        #Redraw the whole GUI (update color codes, plot all curves and so on)
        self.redrawGUI()
        self.changeEnabled(True,['prevRegression','nextRegression'])
        
                        
    def on_menuOpen_activate(self, widget, data=None):
        """ Select and open a Biolector file."""
        
        fileName=None
        
        chooser=gtk.FileChooserDialog("Open File...", self._w['window'],
                                      gtk.FILE_CHOOSER_ACTION_OPEN,
                                      (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, 
                                      gtk.STOCK_OPEN, gtk.RESPONSE_OK))
                                      
        filter=gtk.FileFilter()
        filter.set_name("CSV files")
        filter.add_pattern("*.csv")
        chooser.add_filter(filter)
                
        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            fileName = chooser.get_filename()
            chooser.destroy()
        
        if not fileName is None:
            self._controller.loadFile(fileName)
            
            #Redraw the whole GUI (update color codes, plot all curves and so on)
            self.redrawGUI()
            self.changeEnabled(True,['prevRegression','nextRegression'])
            
    def plotGroups(self,groupIndexes,plotRegression=False):
        """ Plot groups on both upper and lower figure.
        
            The logarithm will be used for the values for the lower figure.
        """
        
        #Plot the data on the upper and lower figures
        values=None
        colors=None
        lines=None
        dashed=None
        intervals=None #Vertical lines to indicate regression interval
        regressions=None #Results structure from fitting
        
        if len(groupIndexes)>0:
            for index in groupIndexes:
                wells=self._controller.getGroupMembers(index)
                if len(wells)>0:
                    #Get which wells to plot and calculate the average values for them
                    I=self._controller.toPlot[wells]
                    wells=wells[I]
                    
                    if len(wells)>1:
                        meanValues=np.mean(self._controller.values[:,wells],1)[:,None]
                    else:
                        if len(wells)>0:
                            meanValues=self._controller.values[:,wells]
                        else:
                            #Nothing to display, continue
                            continue
                    
                    #Add to calculated means
                    gColor=self._controller.getGroupColor(index)[0]
                    if values is None:
                        values=meanValues
                        colors=gColor
                        dashed=np.zeros(1,dtype=bool)
                        if plotRegression is True:
                            intervals=self._controller.getGroupRegressionLimits(index)
                            regressions=self._controller.getGroupRegressionResults([index])
                        
                    else:
                        values=my.cat(values,meanValues,1)
                        colors=my.cat(colors,gColor,0)
                        dashed=my.cat(dashed,np.zeros(1,dtype=bool),0)
                        if plotRegression is True:
                            intervals=my.cat(intervals,self._controller.getGroupRegressionLimits(index),0)
                            regressions=regressions+self._controller.getGroupRegressionResults([index])
                        
                    #Also add the raw values for the members of the group
                    values=my.cat(values,self._controller.values[:,wells],1)
                    colors=my.cat(colors,np.tile(gColor,(len(wells),1)),0)
                    dashed=my.cat(dashed,np.ones(len(wells),dtype=bool),0)
                    
                    #There should not be drawn intervals for members of a group
                    if plotRegression is True:
                        nan=np.array([[np.nan,np.nan]])
                        intervals=my.cat(intervals,np.tile(nan,(len(wells),1)),0)
                        
                        #Or plotted regression lines
                        for i in range(0,len(wells)):
                            regressions.append(None)
            
            #It could be that no wells should be plotted
            if not values is None:
                if values.shape[1]>0:
                    #Plot data on upper figure
                    self.plotData(self._controller.timePoints,values,colors,dashed,intervals,regressions,0)
        
                    #Plot data on lower figure
                    self.plotData(self._controller.timePoints,np.log(values),colors,dashed,intervals,regressions,1)
                
    def plotWells(self,wellIndexes,plotRegression=False):
        """ Plot individual wells on both upper and lower figure.
        
            The logarithm will be used for the values for the lower figure.
        """
        intervals=None
        regressions=None
        
        if len(wellIndexes)>0:
            values=self._controller.values[:,wellIndexes]
            colors=self._controller.colors[wellIndexes,:]
            dashed=np.zeros(len(wellIndexes),dtype=bool)
            if plotRegression is True:
                intervals=self._controller.regressionLimits[wellIndexes,:]
                regressions=self._controller.getRegressionResults(wellIndexes)
            
            #Plot data on upper figure
            self.plotData(self._controller.timePoints,values,colors,dashed,intervals,regressions,0)
            
            #Plot data on lower figure
            self.plotData(self._controller.timePoints,np.log(values),colors,dashed,intervals,regressions,1)
            
    def redrawFigures(self):
        """ Redraw figures."""
        
        self._uCanvas.draw()
        self._lCanvas.draw()
        
    def clearFigures(self):
        """ Clear figures."""
        
        self._uFig.clear()
        self._lFig.clear()
            
    def redrawGUI(self):
        """ Redraw the whole GUI."""
        
        #Color the well labels
        self.colorWells()
        
        #Populate the TreeView with the wells
        self.updateTreeView()
        
        #Clear figures
        self.clearFigures()

        #First plot all groups
        self.plotGroups(np.where(self._controller.toPlotGroups)[0])
        
        #Then add the individual wells
        self.plotWells(np.where(np.logical_and(self._controller.inGroup==False, self._controller.toPlot==True))[0])
        
        #Redraw figures
        self.redrawFigures()
        
    def updateTreeView(self):
        """ Update the TreeView."""
        
        self._w['treeStore'].clear() #Clear everything
        #First add the groups
        if not self._controller.groupLabels is None:
            for i in range(0,len(self._controller.groupLabels)):
                #This is a little ugly, but I use negative indexes to represent groups and positive for wells
                #Note that it's "i-1" for groups to avoid 0 being an index for both wells and groups
                wells=self._controller.getGroupMembers(i)
                
                if self._controller.toPlotGroups[i]==False:
                    color=self._defaultGray
                else:
                    color=self._controller.getGroupColor(i)[1]
                    if color is None:
                        color=self._defaultWhite
                    
                group=self._w['treeStore'].append(None,[self._controller.groupLabels[i],color,self._controller.toPlotGroups[i],'   ',-i-1,True])
                
                #Then add the wells in the group
                for well in wells:
                    if self._controller.toPlot[well]==True:
                        wellcolor=color
                    else:
                        wellcolor=self._defaultGray
                    self._w['treeStore'].append(group,[self._controller.wellLabels[well],wellcolor,self._controller.toPlot[well],'   ',well,True])
                
        #Then add the individual wells
        for i in range(0,len(self._controller.wellLabels)):
            if self._controller.inGroup[i]==False:
                if self._controller.toPlot[i]==True:
                    color=gtk.gdk.Color(self._controller.colors[i,0],self._controller.colors[i,1],self._controller.colors[i,2])
                else:
                    color=self._defaultGray
                self._w['treeStore'].append(None,[self._controller.wellLabels[i],color,self._controller.toPlot[i],'   ',i,True])
    
    def colorWells(self):
        """ Set the colors for the well labels."""
        
        #Go through the groups and color all wells associated to them
        groups=self._controller.toPlotGroups
        if not groups is None:
            for i in range(0,len(groups)):
                wells=self._controller.getGroupMembers(i)
                if len(wells)>0:
                    #Use the color of the first well
                    color=self._controller.colors[wells[0],:]
                
                    for well in wells:
                        if self._controller.toPlot[well]==True:
                            self.setWellColor(well,color)
                        else:
                            self.setWellColor(well)
        
        #Then color wells which do not belong to a group
        I=np.where(self._controller.inGroup==False)
        
        for i in I[0]:
            if self._controller.toPlot[i]==True:
                self.setWellColor(i,self._controller.colors[i,:])
            else:
                self.setWellColor(i)
            
    def setWellColor(self,index,color=None):
        """ Set the color of an individual well.
        
        If color is not defined the default gray is used. Note that color is
        RBGA but that gdk.Color is used by the ListStore"""
        
        I=np.unravel_index(index,(6,8)) #Map from linear index
        
        if not color is None:
            self._w['wellListStore'][1+I[0]][I[1]+8]=gtk.gdk.Color(color[0],color[1],color[2])
        else:
            self._w['wellListStore'][1+I[0]][I[1]+8]=self._defaultGray

    def on_menuAbout_activate(self, widget, data=None):
        """ Display information about the software."""

        about = gtk.AboutDialog()
        about.set_program_name("GetGrowth")
        about.set_version("0.2")
        about.set_copyright("(c) Novo Nordisk A/S")
        about.set_comments("This software is not to be distributed outside of Novo Nordisk. Please contact Rasmus Ågren (RAAG) for any questions or comments.")
        about.run()
        about.destroy()
            
    def on_menuManual_activate(self, widget, data=None):
        """ Open the manual in the default PDF reader."""
        __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
        subprocess.Popen(os.path.join(__location__, 'docs/manual.pdf'),shell=True)
        
    def on_window_destroy(self, widget, data=None):
        """ Quit the main GUI loop."""
        
        gtk.main_quit()
    
    def on_menuQuit_activate(self, widget, data=None):
        """ Close the window."""
        self._w['window'].destroy()
    
    #Callers
    def on_newGroup_clicked(self,widget,data=None):
        """ Catch clicks on the newGroup button."""
        
        #Only respond to the click if some data has been loaded
        if not self._controller.values is None:
            self._controller.newGroup()
            
            #Redraw the GUI. More specific functions should probably be used later
            self.redrawGUI()    
        
    def on_treeViewToggle_toggled(self,widget,data=None):
        """ Catch toggle for rows in the treeView."""
        
        #Get the index for the well/group
        I=self._w['treeStore'][data][4]
        
        #If it's a well
        if I>=0:
            self._controller.toggleToPlot(I)
        else:
            #If it's a group
            self._controller.toggleToPlotGroup(abs(I)-1)
        
        #Redraw the GUI. More specific functions should probably be used later
        self.redrawGUI()
    
    def on_treeViewLabels_edited(self,widget,data=None,newText=None):
        
        I=self._w['treeStore'][data][4]
        
        if I>=0:
            #If it's a well
            self._controller.setWellLabel(I,newText)
        else:
            #If it's a group
            self._controller.setGroupLabel(-1*I-1,newText)
            
        self.updateTreeView()
        
    def on_lCanvas_clicked(self,event):
        """ Catch clicks on the lower figure. This is used for setting intervals for
            the regression.
            
            Limits are set via left (for lower) and right (for upper) mouse clicks.
        """
        
        if self._fittingRegression==True:
            #The "not" in the if statements is to deal with the default value of np.nan()
            if self._currentRegressionIsGroup==True:
                if event.button==1:
                    #Don't do anything if it's higher than the upper limit
                    if not event.xdata>=self._controller._groupRegressionLimits[self._currentRegressionIndex,1]:
                        self._controller._groupRegressionLimits[self._currentRegressionIndex,0]=event.xdata
                #This should correspond to the right button. It's normally "3" but I'm not sure if that's always the case
                else:
                    if not event.xdata<=self._controller._groupRegressionLimits[self._currentRegressionIndex,0]:
                        self._controller._groupRegressionLimits[self._currentRegressionIndex,1]=event.xdata
                 
                #Perform the regression
                self._controller.fitGroup(self._currentRegressionIndex)       
                
            else:
                if event.button==1:
                    #Don't do anything if it's higher than the upper limit
                    if not event.xdata>=self._controller._regressionLimits[self._currentRegressionIndex,1]:
                        self._controller._regressionLimits[self._currentRegressionIndex,0]=event.xdata
                #This should correspond to the right button. It's normally "3" but I'm not sure if that's always the case
                else:
                    #Don't do anything if it's lower than the lower limit
                    if not event.xdata<=self._controller._regressionLimits[self._currentRegressionIndex,0]:
                        self._controller._regressionLimits[self._currentRegressionIndex,1]=event.xdata
                    
                #Perform the regression
                self._controller.fitLine(self._currentRegressionIndex)
             
            self.drawCurrentRegression()
            
    def changeEnabled(self,enable,exclude=['window']):
        """ Enable or disable all GUI components.

            enable:     true if the components should be enabled
            exclude:    list with names to exclude
        """
        
        if exclude is None:
            keys=self._w.keys
        else:
            keys=set(self._w.keys()) - set(exclude)
            
        #Exclude components which cannot be enabled/disabled or which belong to other windows
        keys=set(keys)-set(['wellListStore','treeStore'])
        for key in keys:
            self._w[key].set_sensitive(enable)
                 
    def drawCurrentRegression(self):
        """ Draw the current group/well being used for setting regression limits.
            Also update the regression label.
        """
        #Clear figures
        self.clearFigures()
        if self._currentRegressionIsGroup==True:
            self.plotGroups([self._currentRegressionIndex],True)
            toFit=self._controller._toPlotGroups
            self._w['regressionLabel'].set_label('Group ' + str(sum(toFit[:self._currentRegressionIndex+1])) + '/' + str(sum(toFit)) + ': ' + self._controller.groupLabels[self._currentRegressionIndex])
        else:
            self.plotWells([self._currentRegressionIndex],True)
            toFit=np.bitwise_and(self._controller._toPlot,~self._controller.inGroup)
            self._w['regressionLabel'].set_label('Well ' + str(sum(toFit[:self._currentRegressionIndex+1])) + '/' + str(sum(toFit)) + ': ' + self._controller.wellLabels[self._currentRegressionIndex])
            
        self.redrawFigures()
                
    def on_fitRegression_clicked(self,widget,data=None):
        """ Catch clicks on the fitRegression button."""
        
        if self._fittingRegression==False:
            self._fittingRegression=True
            self.shiftRegressionNumber()
            self.changeEnabled(False,['window','fitRegression','nextRegression','lCanvas','lFig','xScale','regressionLabel'])
            self._w['fitRegression'].set_label('Done')
            self._w['nextRegression'].set_sensitive(True) #It could be false since last fitting
        else:
            self._fittingRegression=False
            self.redrawGUI() #Redraw the whole GUI
            self._w['fitRegression'].set_label('Fit regression lines')
            self._w['regressionLabel'].set_label('')
            self.changeEnabled(True)
            self._w['prevRegression'].set_sensitive(False)
            self._w['nextRegression'].set_sensitive(False)
            
    def on_nextRegression_clicked(self,widget,data=None):
        """ Catch clicks on the nextRegression button."""
        
        #Should always be true
        if self._fittingRegression==True:
            self.shiftRegressionNumber(1)
    
    def on_prevRegression_clicked(self,widget,data=None):
        """ Catch clicks on the prevRegression button."""
        
        #Should always be true
        if self._fittingRegression==True:
            self.shiftRegressionNumber(-1)
    
    def shiftRegressionNumber(self,shift=0):
        """ Shift the regression item number and show the corresponding plots.
        
            shift: -1 for previous, 0 for reset to 0, 1 for next
        """
        
        if not self._controller.toPlotGroups is None:
            nGroups=sum(self._controller.toPlotGroups)
        else:
            nGroups=0
         
        if not self._controller.toPlot is None:
            nWells=sum(np.bitwise_and(self._controller.toPlot,~self._controller.inGroup))
        else:
            nWells=0
        
        #Reset to use the first element
        if shift==0:
            self._currentRegressionNumber=0
            
        if shift==-1:
            self._currentRegressionNumber=np.max((self._currentRegressionNumber-1,0))
        
        if shift==1:
            self._currentRegressionNumber=np.min((self._currentRegressionNumber+1,nGroups+nWells-1))
                    
        #Plot a group
        if nGroups>self._currentRegressionNumber:
            self._currentRegressionIsGroup=True
            gIds=np.where(self._controller.toPlotGroups)[0]
            self._currentRegressionIndex=gIds[self._currentRegressionNumber]
        else:
            self._currentRegressionIsGroup=False
            wIds=np.where(np.bitwise_and(self._controller.toPlot,~self._controller.inGroup))[0]
            self._currentRegressionIndex=wIds[self._currentRegressionNumber-nGroups]
                
        #Enable or disable the buttons
        self._w['prevRegression'].set_sensitive(True)
        self._w['nextRegression'].set_sensitive(True)
        if self._currentRegressionNumber==0:
            self._w['prevRegression'].set_sensitive(False)
        if self._currentRegressionNumber==nGroups+nWells-1:
            self._w['nextRegression'].set_sensitive(False)
        
        self.drawCurrentRegression() 
            
    def on_addToGroup_clicked(self,widget,data=None):
        """ Catch clicks on the addToGroup button."""
        
        if not self._controller.groupLabels is None:
            #Only do this if it's not already adding to a group
            if self._addingToGroup is None:
                #First check if there is a group selected in the treeView
                treeSelection = self._w['treeView'].get_selection()
                (model, pathlist)=treeSelection.get_selected_rows()
                
                #This should currently be only one, but for the future maybe
                for path in pathlist:
                    treeIter=model.get_iter(path)
                    value=model.get_value(treeIter,4)
                    
                    #Only do something if a group was selected
                    if value<0:
                        self._addingToGroup=abs(value)-1
                        self.changeEnabled(False,['window','wellTreeView','addToGroup'])
                        self._w['addToGroup'].set_label('Done')
            else:
                self._addingToGroup=None
                self.changeEnabled(True)
                self._w['addToGroup'].set_label('Add to selected group')           
    
    def on_wellTreeView_button_press_event(self,widget,data=None):
        """ Catch clicks on the wellTreeView.
        
            This is done in a rather inconvenient way since TreeViews are based
            around selecting cells and that leads to them being colored. To avoid
            that the indexes are calculated from the position of the click instead.
            
            This is very unstable and must be done better at some point.
        """
        
        #Only respond to the click if some data has been loaded
        if not self._controller.values is None:
            rect=widget.get_allocation()
            
            #Calculate indexes
            I=np.floor(9*data.x/rect.width)
            J=np.floor(7*data.y/rect.height)

            #If the click was on a well
            if I>0 and J>0:
                #Convert to linear index
                ind=np.int((J-1)*8+I-1)
                
                #Toggle whether to display the well or not or if it should be added to a group
                if self._addingToGroup is None:
                    self._controller.toggleToPlot(ind)
                else:
                    self._controller.addToGroup(self._addingToGroup,ind)
            else:
                #If the click was on a whole row/column
                if I==0 and J==0:
                    self._controller.toggleMultipleToPlot() #Toggle all
                else:
                    if I==0:
                        self._controller.toggleMultipleToPlot(J-1,0) #Toggle row
                    else:
                        self._controller.toggleMultipleToPlot(I-1,1) #Toggle column
            
            #Redraw the GUI. More specific functions should probably be used later
            self.redrawGUI()
        
        #I don't quite know how this works, but by returning True you inactivate the
        #selection of the row
        return True